class SecurityLabelSystems:
    owner_codeset: str = "https://www.icanbwell.com/owner"

    access_codeset: str = "https://www.icanbwell.com/access"

    vendor_codeset: str = "https://www.icanbwell.com/vendor"

    connection_type_codeset: str = "https://www.icanbwell.com/connectionType"

    source_assigning_authority_codeset: str = (
        "https://www.icanbwell.com/sourceAssigningAuthority"
    )
